@@ -17,8 +17,6 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/* $Id: CopyJob.cc,v 1.48 2011/04/29 04:58:27 lav Exp $ */
-
 #include <config.h>
 #include "CopyJob.h"
 #include "ArgV.h"
@@ -40,7 +38,12 @@ int CopyJob::Do()
       return STALL;
    if(c->Error())
    {
-      eprintf("%s: %s\n",op.get(),c->ErrorText());
+      const char *error=c->ErrorText();
+      const char *name=GetDispName();
+      if(!strstr(error,name) && op.ne(name))
+	 error=xstring::cat(name,": ",error,NULL);
+      if(!quiet)
+	 eprintf("%s: %s\n",op.get(),error);
       done=true;
       return MOVED;
    }
@@ -135,7 +138,7 @@ void CopyJob::SetDispName()
 }
 
 CopyJob::CopyJob(FileCopy *c1,const char *name1,const char *op1)
-   : c(c1), name(name1), op(op1)
+   : c(c1), name(name1), op(op1), quiet(false)
 {
    done=false;
    no_status=false;
@@ -158,7 +161,7 @@ const char *CopyJob::FormatBytesTimeRate(off_t bytes,double time_spent)
 		     (long long)bytes,long(time_spent+.5));
       double rate=bytes/time_spent;
       if(rate>=1)
-	 msg.appendf(" (%s)",Speedometer::GetStr(rate).get());
+	 msg.appendf(" (%s)",Speedometer::GetStrProper(rate).get());
       return msg;
    }
    return xstring::format(plural("%lld $#ll#byte|bytes$ transferred",
@@ -177,7 +180,7 @@ CopyJob::~CopyJob()
 
 // CopyJobEnv
 CopyJobEnv::CopyJobEnv(FileAccess *s,ArgV *a,bool cont1)
-   : SessionJob(s)
+   : SessionJob(s), quiet(false)
 {
    args=a;
    args->rewind();
@@ -186,6 +189,7 @@ CopyJobEnv::CopyJobEnv(FileAccess *s,ArgV *a,bool cont1)
    cp=0;
    errors=0;
    count=0;
+   parallel=ResMgr::Query("xfer:parallel",0);
    bytes=0;
    time_spent=0;
    no_status=false;
@@ -202,7 +206,7 @@ int CopyJobEnv::Do()
    int m=STALL;
    if(done)
       return m;
-   if(waiting_num<1)
+   if(waiting_num<parallel)
    {
       NextFile();
       if(waiting_num==0)
@@ -217,16 +221,27 @@ int CopyJobEnv::Do()
    if(j==0)
       return m;
    RemoveWaiting(j);
-   if(j->ExitCode()!=0)
+   if(j->GetLocal())
+   {
+      if(j->Error()) {
+	 // in case of errors, move the backup to original location
+	 j->GetLocal()->revert_backup();
+      } else {
+	 // now we can delete the old file, since there is a new one
+	 j->GetLocal()->remove_backup();
+      }
+   }
+   if(j->Error())
       errors++;
    count++;
    bytes+=j->GetBytesCount();
-   time_spent+=j->GetTimeSpent();
-   Delete(j);
    if(cp==j)
       cp=0;
+   Delete(j);
    if(waiting_num>0 && cp==0)
       cp=(CopyJob*)waiting[0];
+   if(waiting.count()==0)
+      time_spent+=now-transfer_start_ts;
    return MOVED;
 }
 void CopyJobEnv::AddCopier(FileCopy *c,const char *n)
@@ -236,6 +251,9 @@ void CopyJobEnv::AddCopier(FileCopy *c,const char *n)
    if(ascii)
       c->Ascii();
    cp=cj_new?cj_new->New(c,n,op):new CopyJob(c,n,op);
+   cp->Quiet(quiet);
+   if(waiting.count()==0)
+      transfer_start_ts=now;
    AddWaiting(cp);
 }
 void CopyJobEnv::SetCopier(FileCopy *c,const char *n)
@@ -281,7 +299,8 @@ xstring& CopyJobEnv::FormatStatus(xstring& s,int v,const char *prefix)
 
 void CopyJobEnv::SayFinal()
 {
-   printf("%s",FormatFinalWithPrefix(xstring::get_tmp(""),"").get());
+   if(!quiet)
+      printf("%s",FormatFinalWithPrefix(xstring::get_tmp(""),"").get());
 }
 
 int CopyJobEnv::AcceptSig(int sig)
