@@ -167,6 +167,8 @@ void Http::Init()
 
    no_ranges=false;
    seen_ranges_bytes=false;
+   entity_date_set=false;
+   sending_proppatch=false;
 
    no_cache_this=false;
    no_cache=false;
@@ -175,8 +177,6 @@ void Http::Init()
    auth_scheme[0]=auth_scheme[1]=HttpAuth::NONE;
 
    use_propfind_now=true;
-   allprop="";
-   allprop_len=0;
 
    retry_after=0;
 
@@ -237,10 +237,10 @@ void Http::DisconnectLL()
    && !Error() && !H_AUTH_REQ(status_code)) {
       if(last_method && !strcmp(last_method,"POST"))
 	 SetError(FATAL,_("POST method failed"));
-      else if(mode==STORE)
+      else if(ModeIs(STORE))
 	 SetError(STORE_FAILED,0);
    }
-   if(mode==STORE && H_AUTH_REQ(status_code))
+   if(ModeIs(STORE) && H_AUTH_REQ(status_code))
       pos=real_pos=request_pos; // resend all the data again
 
    last_method=0;
@@ -271,6 +271,7 @@ void Http::ResetRequestData()
    propfind=0;
    inflate=0;
    seen_ranges_bytes=false;
+   entity_date_set=false;
 }
 
 void Http::Close()
@@ -280,7 +281,7 @@ void Http::Close()
    if(conn && conn->recv_buf)
       conn->recv_buf->Roll();	// try to read any remaining data
    if(conn && keep_alive && (keep_alive_max>0 || keep_alive_max==-1)
-   && mode!=STORE && !conn->recv_buf->Eof() && (state==RECEIVING_BODY || state==DONE))
+   && !ModeIs(STORE) && !conn->recv_buf->Eof() && (state==RECEIVING_BODY || state==DONE))
    {
       conn->recv_buf->Resume();
       conn->recv_buf->Roll();
@@ -314,19 +315,25 @@ void Http::Close()
    use_propfind_now=QueryBool("use-propfind",hostname);
    special=HTTP_NONE;
    special_data.set(0);
+   sending_proppatch=false;
    super::Close();
 }
 
+void Http::Send(const xstring& str)
+{
+   if(str.length()==0)
+      return;
+   LogSend(5,str);
+   conn->send_buf->Put(str);
+}
+
 void Http::Send(const char *format,...)
 {
    va_list va;
    va_start(va,format);
    xstring& str=xstring::vformat(format,va);
    va_end(va);
-   if(str.length()==0)
-      return;
-   LogSend(5,str);
-   conn->send_buf->Put(str);
+   Send(str);
 }
 
 void Http::Send(const HttpHeader *hdr)
@@ -582,16 +589,46 @@ void Http::SendPropfind(const xstring& efile,int depth)
 {
    SendMethod("PROPFIND",efile);
    Send("Depth: %d\r\n",depth);
-   if(allprop_len>0)
+   if(allprop.length()>0)
    {
       Send("Content-Type: text/xml\r\n");
-      Send("Content-Length: %d\r\n",allprop_len);
+      Send("Content-Length: %d\r\n",int(allprop.length()));
    }
 }
 void Http::SendPropfindBody()
 {
-   if(allprop_len>0)
-      Send("%s",allprop);
+   Send(allprop);
+}
+
+const xstring& Http::FormatLastModified(time_t lm)
+{
+   static const char weekday_names[][4]={
+      "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
+   };
+   const struct tm *t=gmtime(&lm);
+   return xstring::format("%s, %2d %s %04d %02d:%02d:%02d GMT",
+      weekday_names[t->tm_wday],t->tm_mday,month_names[t->tm_mon],
+      t->tm_year+1900,t->tm_hour,t->tm_min,t->tm_sec);
+}
+
+void Http::SendProppatch(const xstring& efile)
+{
+   SendMethod("PROPPATCH",efile);
+   xstring prop(
+      "<?xml version=\"1.0\" ?>"
+      "<propertyupdate xmlns=\"DAV:\">"
+	 "<set>"
+	    "<prop>"
+	       "<getlastmodified>");
+   prop.append(FormatLastModified(entity_date)).append(
+	       "</getlastmodified>"
+	    "</prop>"
+	 "</set>"
+      "</propertyupdate>");
+   Send("Content-Type: text/xml\r\n");
+   Send("Content-Length: %d\r\n",int(prop.length()));
+   Send("\r\n");
+   Send(prop);
 }
 
 void Http::SendRequest(const char *connection,const char *f)
@@ -666,7 +703,7 @@ void Http::SendRequest(const char *connection,const char *f)
 
    if(pos==0)
       real_pos=0;
-   if(mode==STORE)    // can't seek before writing
+   if(ModeIs(STORE))    // can't seek before writing
       real_pos=pos;
 
 #ifdef DEBUG_MP_LIST
@@ -719,6 +756,10 @@ void Http::SendRequest(const char *connection,const char *f)
       break;
 
    case STORE:
+      if(sending_proppatch) {
+	 SendProppatch(efile);
+	 break;
+      }
       if(hftp || strcasecmp(Query("put-method",hostname),"POST"))
 	 SendMethod("PUT",efile);
       else
@@ -746,14 +787,7 @@ void Http::SendRequest(const char *connection,const char *f)
       }
       if(entity_date!=NO_DATE)
       {
-	 static const char weekday_names[][4]={
-	    "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
-	 };
-	 const struct tm *t=gmtime(&entity_date);
-	 const char *d=xstring::format("%s, %2d %s %04d %02d:%02d:%02d GMT",
-	    weekday_names[t->tm_wday],t->tm_mday,month_names[t->tm_mon],
-	    t->tm_year+1900,t->tm_hour,t->tm_min,t->tm_sec);
-	 Send("Last-Modified: %s\r\n",d);
+	 Send("Last-Modified: %s\r\n",FormatLastModified(entity_date).get());
 	 Send("X-OC-MTime: %ld\r\n",(long)entity_date);	 // for OwnCloud
       }
       break;
@@ -812,7 +846,7 @@ void Http::SendRequest(const char *connection,const char *f)
    case RENAME:
       {
 	 SendMethod("MOVE",efile);
-	 Send("Destination: %s\r\n",GetFileURL(file1));
+	 Send("Destination: %s\r\n",GetFileURL(file1).get());
       }
    }
    if(proxy && !https)
@@ -823,7 +857,7 @@ void Http::SendRequest(const char *connection,const char *f)
    SendCacheControl();
    if(mode==ARRAY_INFO && !use_head)
       connection="close";
-   else if(mode!=STORE)
+   else if(!ModeIs(STORE))
       connection="keep-alive";
    if(mode!=ARRAY_INFO || connection)
       Send("Connection: %s\r\n",connection?connection:"close");
@@ -986,7 +1020,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 last=fsize-first-1;
       if(body_size<0)
 	 body_size=last-first+1;
-      if(mode!=STORE && mode!=MAKE_DIR)
+      if(!ModeIs(STORE) && !ModeIs(MAKE_DIR))
 	 entity_size=fsize;
       if(opt_size && H_2XX(status_code))
 	 *opt_size=fsize;
@@ -1101,6 +1135,10 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       NewAuth(value,HttpAuth::PROXY,proxy_user,proxy_pass);
       return;
    }
+   case_hh("X-OC-MTime",'X') {
+      if(!strcasecmp(value,"accepted"))
+	 entity_date_set=true;
+   }
    default:
       break;
    }
@@ -1237,7 +1275,7 @@ int Http::Do()
    case DISCONNECTED:
       if(mode==CLOSED || !hostname)
 	 return m;
-      if(mode==STORE && pos>0 && entity_size>=0 && pos>=entity_size)
+      if(ModeIs(STORE) && pos>0 && entity_size>=0 && pos>=entity_size)
       {
 	 state=DONE;
 	 return MOVED;
@@ -1456,7 +1494,7 @@ int Http::Do()
 	 return MOVED;
       }
       ExpandTildeInCWD();
-      if(mode==STORE && pos>0 && entity_size>=0 && pos>=entity_size)
+      if(ModeIs(STORE) && pos>0 && entity_size>=0 && pos>=entity_size)
       {
 	 state=DONE;
 	 return MOVED;
@@ -1477,13 +1515,13 @@ int Http::Do()
 
       state=RECEIVING_HEADER;
       m=MOVED;
-      if(mode==STORE)
+      if(ModeIs(STORE))
 	 rate_limit=new RateLimit(hostname);
 
    case RECEIVING_HEADER:
       if(conn->send_buf->Error() || conn->recv_buf->Error())
       {
-	 if((mode==STORE || special) && status_code && !H_2XX(status_code))
+	 if((ModeIs(STORE) || special) && status_code && !H_2XX(status_code))
 	    goto pre_RECEIVING_BODY;   // assume error.
       handle_buf_error:
 	 if(conn->send_buf->Error())
@@ -1584,13 +1622,19 @@ int Http::Do()
 		  ProceedArrayInfo();
 		  return MOVED;
 	       }
-	       else if(mode==STORE || mode==MAKE_DIR)
+	       else if(ModeIs(STORE) || ModeIs(MAKE_DIR) || sending_proppatch)
 	       {
-		  if((sent_eot || pos==entity_size) && H_2XX(status_code))
+		  if((sent_eot || pos==entity_size || sending_proppatch) && H_2XX(status_code))
 		  {
 		     state=DONE;
 		     Disconnect();
 		     state=DONE;
+		     if(ModeIs(STORE) && entity_date!=NO_DATE && !entity_date_set
+		     && use_propfind_now) {
+			// send PROPPATCH in a separate request.
+			sending_proppatch=true;
+			state=DISCONNECTED;
+		     }
 		     return MOVED;
 		  }
 		  if(H_2XX(status_code))
@@ -1624,7 +1668,7 @@ int Http::Do()
 		  proto_version=0x09;
 		  status_code=H_Ok;
 		  LogError(0,_("Could not parse HTTP status line"));
-		  if(mode==STORE)
+		  if(ModeIs(STORE))
 		  {
 		     state=DONE;
 		     Disconnect();
@@ -1696,7 +1740,7 @@ int Http::Do()
 	 }
       }
 
-      if(mode==STORE && (!status || H_CONTINUE(status_code)) && !sent_eot)
+      if(ModeIs(STORE) && (!status || H_CONTINUE(status_code)) && !sent_eot)
 	 Block(conn->sock,POLLOUT);
 
       return m;
@@ -1813,7 +1857,7 @@ int Http::Do()
       }
       // sometimes it's possible to derive entity size from body size.
       if(entity_size==NO_SIZE && body_size!=NO_SIZE
-      && pos==0 && mode!=STORE && mode!=MAKE_DIR && !inflate) {
+      && pos==0 && !ModeIs(STORE) && !ModeIs(MAKE_DIR) && !inflate) {
 	 entity_size=body_size;
 	 if(opt_size && H_2XX(status_code))
 	    *opt_size=body_size;
@@ -1823,7 +1867,7 @@ int Http::Do()
       rate_limit=new RateLimit(hostname);
       if(real_pos<0) // assume Range: did not work
       {
-	 if(mode!=STORE && mode!=MAKE_DIR && body_size>=0)
+	 if(!ModeIs(STORE) && !ModeIs(MAKE_DIR) && body_size>=0)
 	 {
 	    entity_size=body_size;
 	    if(opt_size && H_2XX(status_code))
@@ -2165,14 +2209,14 @@ int Http::Done()
 
 int Http::Buffered()
 {
-   if(mode!=STORE || !conn || !conn->send_buf)
+   if(!ModeIs(STORE) || !conn || !conn->send_buf)
       return 0;
    return conn->send_buf->Size()+SocketBuffered(conn->sock);
 }
 
 int Http::Write(const void *buf,int size)
 {
-   if(mode!=STORE)
+   if(!ModeIs(STORE))
       return(0);
 
    Resume();
@@ -2218,7 +2262,7 @@ int Http::SendEOT()
       return OK;
    if(Error())
       return(error_code);
-   if(mode==STORE)
+   if(ModeIs(STORE))
    {
       if(state==RECEIVING_HEADER && conn->send_buf->Size()==0)
       {
@@ -2260,7 +2304,7 @@ const char *Http::CurrentStatus()
    case CONNECTED:
       return(_("Connection idle"));
    case RECEIVING_HEADER:
-      if(mode==STORE && !sent_eot && !status)
+      if(ModeIs(STORE) && !sent_eot && !status)
 	 return(_("Sending data"));
       if(tunnel_state==TUNNEL_WAITING)
 	 return(_("Connecting..."));
@@ -2315,14 +2359,15 @@ void Http::Reconfig(const char *name)
    use_propfind_now=(use_propfind_now && QueryBool("use-propfind",c));
    no_ranges=(no_ranges || !QueryBool("use-range",hostname));
 
-   allprop=	// PROPFIND request
-      "<?xml version=\"1.0\" ?>"
-      "<propfind xmlns=\"DAV:\">"
-        "<allprop/>"
-      "</propfind>\r\n";
-   if(!QueryBool("use-allprop",c))
-      allprop="";
-   allprop_len=strlen(allprop);
+   if(QueryBool("use-allprop",c)) {
+      allprop.set(   // PROPFIND request
+	 "<?xml version=\"1.0\" ?>"
+	 "<propfind xmlns=\"DAV:\">"
+	   "<allprop/>"
+	 "</propfind>\r\n");
+   } else {
+      allprop.unset();
+   }
 
    if(!user || !pass) {
       // get auth info from http:authorization setting
