@@ -269,15 +269,19 @@ void  MirrorJob::HandleFile(FileInfo *file)
 
    // TODO: get rid of local hacks.
 
+   const char *dst_name=file->name;
+   if(FlagSet(TARGET_FLAT))
+      dst_name=basename_ptr(dst_name);
+
    // dir_name returns pointer to static data - need to dup it.
    const char *source_name=dir_file(source_dir,file->name);
    source_name=alloca_strdup(source_name);
-   const char *target_name=dir_file(target_dir,file->name);
+   const char *target_name=dir_file(target_dir,dst_name);
    target_name=alloca_strdup(target_name);
 
    const char *source_name_rel=dir_file(source_relative_dir,file->name);
    source_name_rel=alloca_strdup(source_name_rel);
-   const char *target_name_rel=dir_file(target_relative_dir,file->name);
+   const char *target_name_rel=dir_file(target_relative_dir,dst_name);
    target_name_rel=alloca_strdup(target_name_rel);
 
    FileInfo::type filetype=FileInfo::NORMAL;
@@ -373,7 +377,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    if(remove_source_files)
 	       args.Append("-E");
 	    args.Append("-O");
-	    args.Append(target_is_local?target_dir.get()
+	    args.Append(target_is_local?target_dir
 			:target_session->GetConnectURL());
 	    args.Append(source_session->GetFileURL(file->name));
 	    xstring_ca cmd(args.CombineQuoted());
@@ -392,7 +396,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 if(target_is_local)
 	    dst_peer=new FileCopyPeerFDStream(new FileStream(target_name,O_WRONLY|O_CREAT|(cont_this?0:O_TRUNC)),FileCopyPeer::PUT);
 	 else
-	    dst_peer=new FileCopyPeerFA(target_session->Clone(),file->name,FA::STORE);
+	    dst_peer=new FileCopyPeerFA(target_session->Clone(),dst_name,FA::STORE);
 
 	 FileCopy *c=FileCopy::New(src_peer,dst_peer,cont_this);
 	 if(remove_source_files)
@@ -417,8 +421,15 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 if(recursion_mode==RECURSION_NEVER || FlagSet(NO_RECURSION))
 	    goto skip;
 
-	 bool create_target_dir=true;
+	 bool create_target_subdir=true;
 	 const FileInfo *old=0;
+
+	 if(FlagSet(TARGET_FLAT)) {
+	    create_target_subdir=false;
+	    target_name=target_dir;
+	    goto do_submirror;
+	 }
+
 	 if(target_set)
 	    old=target_set->FindByName(file->name);
 	 if(!old)
@@ -431,7 +442,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 }
 	 else if(old->TypeIs(old->DIRECTORY))
 	 {
-	    create_target_dir=false;
+	    create_target_subdir=false;
 	 }
 	 if(target_is_local && !script_only)
 	 {
@@ -443,7 +454,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 		  // only if not enabled as chmod can clear sgid flags on directories
 		  if(st.st_mode!=(st.st_mode|0700))
 		     chmod(target_name,st.st_mode|0700);
-		  create_target_dir=false;
+		  create_target_subdir=false;
 	       }
 	       else
 	       {
@@ -453,11 +464,12 @@ void  MirrorJob::HandleFile(FileInfo *file)
 		     eprintf("mirror: remove(%s): %s\n",target_name,strerror(errno));
 		     goto skip;
 		  }
-		  create_target_dir=true;
+		  create_target_subdir=true;
 	       }
 	    }
 	 }
 
+      do_submirror:
 	 // launch sub-mirror
 	 MirrorJob *mj=new MirrorJob(this,
 	    source_session->Clone(),target_session->Clone(),
@@ -465,35 +477,10 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 AddWaiting(mj);
 	 mj->cmdline.vset("\\mirror `",source_name_rel,"'",NULL);
 
-	 // inherit flags and other things
-	 mj->SetFlags(flags,1);
-	 mj->UseCache(use_cache);
-
-	 mj->SetExclude(exclude);
-
 	 mj->source_relative_dir.set(source_name_rel);
 	 mj->target_relative_dir.set(target_name_rel);
 
-	 mj->verbose_report=verbose_report;
-	 mj->newer_than=newer_than;
-	 mj->older_than=older_than;
-	 mj->size_range=size_range;
-	 mj->parallel=parallel;
-	 mj->pget_n=pget_n;
-	 mj->pget_minchunk=pget_minchunk;
-	 mj->remove_source_files=remove_source_files;
-	 mj->remove_source_dirs=remove_source_dirs;
-	 mj->skip_noaccess=skip_noaccess;
-	 mj->create_target_dir=create_target_dir;
-	 mj->no_target_dir=no_target_dir;
-	 mj->recursion_mode=recursion_mode;
-
-	 mj->script=script;
-	 mj->script_needs_closing=false;
-	 mj->script_name.set(script_name);
-	 mj->script_only=script_only;
-
-	 mj->max_error_count=max_error_count;
+	 mj->create_target_dir=create_target_subdir;
 
 	 if(verbose_report>=3) {
 	    if(FlagSet(SCAN_ALL_FIRST))
@@ -591,8 +578,11 @@ skip:
    return;
 }
 
-void  MirrorJob::InitSets(const FileSet *source,const FileSet *dest)
+void  MirrorJob::InitSets(Ref<FileSet>& source,const FileSet *dest)
 {
+   if(FlagSet(TARGET_FLAT) && !parent_mirror && dest)
+      source->Sort(FileSet::BYNAME_FLAT);
+
    source->Count(NULL,&stats.tot_files,&stats.tot_symlinks,&stats.tot_files);
 
    to_rm=new FileSet(dest);
@@ -659,6 +649,15 @@ void  MirrorJob::InitSets(const FileSet *source,const FileSet *dest)
    if(!(flags&DELETE))
       to_transfer->SubtractAny(to_rm_mismatched);
 
+   if(FlagSet(TARGET_FLAT) && !parent_mirror && dest) {
+      source->Unsort();
+      to_transfer->UnsortFlat();
+      to_transfer->SubtractDirs();
+      same->UnsortFlat();
+      to_mkdir->Empty();
+      new_files_set->UnsortFlat();
+   }
+
    const char *sort_by=ResMgr::Query("mirror:sort-by",0);
    bool desc=strstr(sort_by,"-desc");
    if(!strncmp(sort_by,"name",4))
@@ -753,7 +752,7 @@ void MirrorJob::HandleChdir(FileAccessRef& session, int &redirections)
       {
 	 // create target dir even if failed to cd to source dir.
 	 if(script)
-	    fprintf(script,"mkdir %s\n",target_session->GetFileURL(target_dir));
+	    fprintf(script,"mkdir %s\n",target_session->GetFileURL(target_dir).get());
 	 if(!script_only)
 	 {
 	    ArgV *a=new ArgV("mkdir");
@@ -949,7 +948,9 @@ int   MirrorJob::Do()
       m=MOVED;
       if(!source_set)
 	 HandleListInfoCreation(source_session,source_list_info,source_relative_dir);
-      if(!target_set && !create_target_dir && (!FlagSet(DEPTH_FIRST) || FlagSet(ONLY_EXISTING)))
+      if(!target_set && !create_target_dir
+      && (!FlagSet(DEPTH_FIRST) || FlagSet(ONLY_EXISTING))
+      && !(FlagSet(TARGET_FLAT) && parent_mirror))
 	 HandleListInfoCreation(target_session,target_list_info,target_relative_dir);
       if(state!=GETTING_LIST_INFO)
       {
@@ -1055,7 +1056,7 @@ int   MirrorJob::Do()
 	 if(!file->TypeIs(file->DIRECTORY))
 	    continue;
 	 if(script)
-	    fprintf(script,"mkdir %s\n",target_session->GetFileURL(file->name));
+	    fprintf(script,"mkdir %s\n",target_session->GetFileURL(file->name).get());
 	 if(!script_only)
 	 {
 	    ArgV *a=new ArgV("mkdir");
@@ -1206,6 +1207,8 @@ int   MirrorJob::Do()
 	 goto pre_FINISHING_FIX_LOCAL;
 
       to_transfer->rewind();
+      if(FlagSet(TARGET_FLAT))
+	 to_transfer->Sort(FileSet::BYNAME_FLAT);
       set_state(TARGET_CHMOD);
       m=MOVED;
       /*fallthrough*/
@@ -1268,13 +1271,15 @@ int   MirrorJob::Do()
    pre_FINISHING_FIX_LOCAL:
       if(target_is_local && !script_only)     // FIXME
       {
-	 to_transfer->LocalUtime(target_dir,/*only_dirs=*/true);
+	 const bool flat=FlagSet(TARGET_FLAT);
+	 to_transfer->Sort(FileSet::BYNAME_FLAT);
+	 to_transfer->LocalUtime(target_dir,/*only_dirs=*/true,flat);
 	 if(flags&ALLOW_CHOWN)
-	    to_transfer->LocalChown(target_dir);
+	    to_transfer->LocalChown(target_dir,flat);
 	 if(!FlagSet(NO_PERMS) && same)
-	    same->LocalChmod(target_dir,get_mode_mask());
+	    same->LocalChmod(target_dir,get_mode_mask(),flat);
 	 if(FlagSet(ALLOW_CHOWN) && same)
-	    same->LocalChown(target_dir);
+	    same->LocalChown(target_dir,flat);
       }
       if(remove_source_files && (same || to_rm_src))
 	 goto pre_SOURCE_REMOVING_SAME;
@@ -1472,6 +1477,32 @@ MirrorJob::MirrorJob(MirrorJob *parent,
       // get file sets and start transfers.
       // See also comment at MirrorJob::MirrorStarted().
       root_transfer_count=parallel_dirs?1:1024;
+
+      // inherit flags and other things
+      SetFlags(parent->flags,1);
+      UseCache(parent->use_cache);
+
+      SetExclude(parent->exclude);
+
+      verbose_report=parent->verbose_report;
+      newer_than=parent->newer_than;
+      older_than=parent->older_than;
+      size_range=parent->size_range;
+      parallel=parent->parallel;
+      pget_n=parent->pget_n;
+      pget_minchunk=parent->pget_minchunk;
+      remove_source_files=parent->remove_source_files;
+      remove_source_dirs=parent->remove_source_dirs;
+      skip_noaccess=parent->skip_noaccess;
+      no_target_dir=parent->no_target_dir;
+      recursion_mode=parent->recursion_mode;
+
+      script=parent->script;
+      script_needs_closing=false;
+      script_name.set(parent->script_name);
+      script_only=parent->script_only;
+
+      max_error_count=parent->max_error_count;
    }
    MirrorStarted();
 }
@@ -1763,6 +1794,7 @@ CMD(mirror)
       OPT_RECURSION,
       OPT_UPLOAD_OLDER,
       OPT_TRANSFER_ALL,
+      OPT_TARGET_FLAT,
    };
    static const struct option mirror_opts[]=
    {
@@ -1823,6 +1855,7 @@ CMD(mirror)
       {"recursion",required_argument,0,OPT_RECURSION},
       {"upload-older",no_argument,0,OPT_UPLOAD_OLDER},
       {"transfer-all",no_argument,0,OPT_TRANSFER_ALL},
+      {"flat",no_argument,0,OPT_TARGET_FLAT},
       {0}
    };
 
@@ -2070,6 +2103,9 @@ CMD(mirror)
       case(OPT_TRANSFER_ALL):
 	 flags|=MirrorJob::TRANSFER_ALL;
 	 break;
+      case(OPT_TARGET_FLAT):
+	 flags|=MirrorJob::TARGET_FLAT|MirrorJob::SCAN_ALL_FIRST|MirrorJob::DEPTH_FIRST;
+	 break;
       case('?'):
 	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
       no_job:
