@@ -149,14 +149,11 @@ int SFtp::Do()
       const char *prog=Query("connect-program",hostname);
       if(!prog || !prog[0])
 	 prog="ssh -a -x";
-      char *a=alloca_strdup(prog);
-      ArgV *cmd=new ArgV;
-      for(a=strtok(a," "); a; a=strtok(0," "))
-	 cmd->Add(a);
+      ArgV args;
       if(!strchr(init,'/'))
       {
 	 if(init[0])
-	    cmd->Add("-s");   // run ssh2 subsystem
+	    args.Add("-s");   // run ssh2 subsystem
 	 // sftpd does not have a greeting
 	 received_greeting=true;
       }
@@ -164,20 +161,21 @@ int SFtp::Do()
 	 init=xstring::cat("echo SFTP: >&2;",init,NULL);
       if(user)
       {
-	 cmd->Add("-l");
-	 cmd->Add(user);
+	 args.Add("-l");
+	 args.Add(user);
       }
       if(portname)
       {
-	 cmd->Add("-p");
-	 cmd->Add(portname);
+	 args.Add("-p");
+	 args.Add(portname);
       }
-      cmd->Add(hostname);
+      args.Add(hostname);
       if(init[0])
-	 cmd->Add(init);
-      xstring_ca cmd_str(cmd->Combine(0));
+	 args.Add(init);
+      xstring_ca cmd_q(args.CombineShellQuoted(0));
+      xstring& cmd_str=xstring::cat(prog," ",cmd_q.get(),NULL);
       LogNote(9,"%s (%s)",_("Running connect program"),cmd_str.get());
-      ssh=new PtyShell(cmd);
+      ssh=new PtyShell(cmd_str);
       ssh->UsePipes();
       state=CONNECTING;
       timeout_timer.Reset();
@@ -2281,6 +2279,7 @@ int SFtpListInfo::Do()
 	 {
 	    if(file->filetype==file->SYMLINK && follow_symlinks)
 	    {
+	       file->filetype=file->UNKNOWN;
 	       file->defined &= ~(file->SIZE|file->DATE|file->SYMLINK_DEF|file->MODE|file->TYPE|file->USER|file->GROUP);
 	       file->Need(file->SIZE|file->DATE|file->MODE|file->TYPE|file->USER|file->GROUP);
 	    }
