@@ -55,6 +55,7 @@
 #include "OutputJob.h"
 #include "echoJob.h"
 #include "EditJob.h"
+#include "mmvJob.h"
 
 #include "misc.h"
 #include "alias.h"
@@ -88,7 +89,7 @@ CMD(module); CMD(mrm); CMD(mv); CMD(open); CMD(pwd); CMD(queue);
 CMD(repeat); CMD(rm); CMD(scache); CMD(set); CMD(shell); CMD(sleep);
 CMD(slot); CMD(source); CMD(subsh); CMD(suspend); CMD(tasks); CMD(torrent);
 CMD(user); CMD(ver); CMD(wait); CMD(empty); CMD(notempty); CMD(true);
-CMD(false);
+CMD(false); CMD(mmv);
 
 #define HELP_IN_MODULE "m"
 #define ALIAS_FOR(cmd) cmd_##cmd,0,#cmd
@@ -340,8 +341,11 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 " -O <base> specifies base directory or URL where files should be placed\n")},
    {"mrm",     cmd_mrm,    N_("mrm <files>"),
 	 N_("Removes specified files with wildcard expansion\n")},
-   {"mv",      cmd_mv,	    N_("mv <file1> <file2>"),
+   {"mv",      cmd_mv,	   N_("mv <file1> <file2>"),
 	 N_("Rename <file1> to <file2>\n")},
+   {"mmv",      cmd_mmv,   N_("mmv [OPTS] <files> <target-dir>"),
+	 N_("Move <files> to <target-directory> with wildcard expansion\n"
+	 " -O <dir>  specifies the target directory (alternative way)\n")},
    {"nlist",   cmd_ls,     N_("[re]nlist [<args>]"),
 	 N_("List remote file names.\n"
 	 "By default, nlist output is cached, to see new listing use `renlist' or\n"
@@ -1153,6 +1157,7 @@ Job *CmdExec::builtin_glob()
    int opt;
    GlobURL::type_select glob_type=GlobURL::FILES_ONLY;
    const char *cmd=0;
+   bool nullglob=false;
 
    static struct option glob_options[]=
    {
@@ -1176,9 +1181,11 @@ Job *CmdExec::builtin_glob()
 	 break;
       case('e'):
 	 cmd=".notempty";
+	 nullglob=true;
 	 break;
       case('E'):
 	 cmd=".empty";
+	 nullglob=true;
 	 break;
       case('?'):
 	 eprintf(_("Try `help %s' for more information.\n"),op);
@@ -1192,6 +1199,7 @@ Job *CmdExec::builtin_glob()
    if(args->count()<2)
    {
       eprintf(_("Usage: %s [OPTS] command args...\n"),op);
+      RevertToSavedSession();
       return 0;
    }
    assert(args_glob==0 && glob==0);
@@ -1203,10 +1211,12 @@ Job *CmdExec::builtin_glob()
    {
       args_glob=0;
       args->rewind();
+      RevertToSavedSession();
       return cmd_command(this);
    }
    glob=new GlobURL(session,pat,glob_type);
-   RevertToSavedSession();
+   if(nullglob)
+      glob->NullGlob();
    builtin=BUILTIN_GLOB;
    return this;
 }
@@ -1704,12 +1714,28 @@ CMD(cat)
 
 CMD(get)
 {
+   static struct option get_options[] = {
+      {"continue",no_argument,0,'c'},
+      {"Remove-source-files",no_argument,0,'E'},
+      {"remove-target",no_argument,0,'e'},
+      {"ascii",no_argument,0,'a'},
+      {"target-directory",required_argument,0,'O'},
+      {"destination-directory",required_argument,0,'O'},
+      {"quiet",no_argument,0,'q'},
+      {"parallel",optional_argument,0,'P'},
+      {"use-pget-n",optional_argument,0,'n'},
+      {"glob",no_argument,0,256+'g'},
+      {"reverse",no_argument,0,256+'R'},
+      {0}
+   };
+   const char *opts="+cEeaO:qP";
+
    int opt;
    bool cont=false;
-   const char *opts="+cEeuaO:q";
    const char *op=args->a0();
-   ArgV	 *get_args=new ArgV(op);
+   Ref<ArgV> get_args(new ArgV(op));
    int n_conn=1;
+   int parallel=0;
    bool del=false;
    bool del_target=false;
    bool ascii=false;
@@ -1717,40 +1743,35 @@ CMD(get)
    bool make_dirs=false;
    bool reverse=false;
    bool quiet=false;
-   xstring_c output_dir;
+   const char *output_dir=0;
 
    if(!strncmp(op,"re",2))
    {
       cont=true;
-      opts="+EuaO:q";
+      opts="+EaO:qP:";
    }
    if(!strcmp(op,"pget"))
    {
-      opts="+n:ceuO:q";
+      opts="+n:ceO:q";
       n_conn=0; // default, which means to take pget:default-n
    }
    else if(!strcmp(op,"put") || !strcmp(op,"reput"))
    {
       reverse=true;
    }
-   else if(!strcmp(op,"mget"))
+   else if(!strcmp(op,"mget") || !strcmp(op,"mput"))
    {
       glob=true;
-      opts="cEeadO:q";
-   }
-   else if(!strcmp(op,"mput"))
-   {
-      glob=true;
-      opts="cEeadO:q";
-      reverse=true;
+      opts="cEeadO:qP:";
+      reverse=(op[1]=='p');
    }
    if(!reverse)
    {
       const char *od=ResMgr::Query("xfer:destination-directory",session->GetHostName());
       if(od && *od)
-	 output_dir.set(od);
+	 output_dir=od;
    }
-   while((opt=args->getopt(opts))!=EOF)
+   while((opt=args->getopt_long(opts,get_options))!=EOF)
    {
       switch(opt)
       {
@@ -1758,12 +1779,15 @@ CMD(get)
 	 cont=true;
 	 break;
       case('n'):
-	 if(!isdigit((unsigned char)optarg[0]))
-	 {
-	    eprintf(_("%s: -n: Number expected. "),op);
-	    goto err;
-	 }
-	 n_conn=atoi(optarg);
+	 if(optarg) {
+	    if(!isdigit((unsigned char)optarg[0]))
+	    {
+	       eprintf(_("%s: %s: Number expected. "),"-n",op);
+	       goto err;
+	    }
+	    n_conn=atoi(optarg);
+	 } else
+	    n_conn=3;
 	 break;
       case('E'):
 	 del=true;
@@ -1778,18 +1802,39 @@ CMD(get)
 	 make_dirs=true;
 	 break;
       case('O'):
-	 output_dir.set(optarg);
+	 output_dir=optarg;
 	 break;
       case('q'):
 	 quiet=true;
 	 break;
+      case('P'):
+	 if(optarg) {
+	    if(!isdigit((unsigned char)optarg[0]))
+	    {
+	       eprintf(_("%s: %s: Number expected. "),"-P",op);
+	       goto err;
+	    }
+	    parallel=atoi(optarg);
+	 } else
+	    parallel=3;
+	 break;
+      case(256+'R'):
+	 reverse=!reverse;
+	 break;
+      case(256+'g'):
+	 glob=true;
+	 break;
       case('?'):
       err:
 	 eprintf(_("Try `help %s' for more information.\n"),op);
-	 delete get_args;
 	 return 0;
       }
    }
+   if(cont && del_target) {
+      eprintf(_("%s: --continue conflicts with --remove-target.\n"),op);
+      return 0;
+   }
+   JobRef<GetJob> j;
    if(glob)
    {
       if(args->getcurr()==0)
@@ -1799,56 +1844,49 @@ CMD(get)
 	 eprintf(_("File name missed. "));
 	 goto err;
       }
-      delete get_args;
-      // remove options
-      while(args->getindex()>1)
-	 args->delarg(1);
-      mgetJob *j=new mgetJob(session->Clone(),args.borrow(),cont,make_dirs);
-      if(reverse)
-	 j->Reverse();
-      if(del)
-	 j->DeleteFiles();
-      if(ascii)
-	 j->Ascii();
+      mgetJob *mj=new mgetJob(session->Clone(),args,cont,make_dirs);
       if(output_dir)
-	 j->OutputDir(output_dir.borrow());
-      j->Quiet(quiet);
-      return j;
+	 mj->OutputDir(output_dir);
+      j=mj;
    }
-   args->back();
-   const char *a=args->getnext();
-   if(a==0)
-      goto file_name_missed;
-   while(a)
+   else
    {
-      const char *src=a;
-      const char *dst=0;
-      a=args->getnext();
-      if(a && !strcmp(a,"-o"))
+      args->back();
+      const char *a=args->getnext();
+      if(a==0)
+	 goto file_name_missed;
+      while(a)
       {
-	 dst=args->getnext();
+	 const char *src=a;
+	 const char *dst=0;
 	 a=args->getnext();
+	 if(a && !strcmp(a,"-o"))
+	 {
+	    dst=args->getnext();
+	    a=args->getnext();
+	 }
+	 if(reverse)
+	    src=expand_home_relative(src);
+	 dst=output_file_name(src,dst,!reverse,output_dir,false);
+	 get_args->Append(src);
+	 get_args->Append(dst);
       }
-      if(reverse)
-	 src=expand_home_relative(src);
-      dst=output_file_name(src,dst,!reverse,output_dir,false);
-      get_args->Append(src);
-      get_args->Append(dst);
+      j=new GetJob(session->Clone(),get_args.borrow(),cont);
    }
-
-   GetJob *j=new GetJob(session->Clone(),get_args,cont);
+   if(reverse)
+      j->Reverse();
    if(del)
       j->DeleteFiles();
    if(del_target)
       j->RemoveTargetFirst();
    if(ascii)
       j->Ascii();
-   if(reverse)
-      j->Reverse();
    if(n_conn!=1)
       j->SetCopyJobCreator(new pCopyJobCreator(n_conn));
+   if(parallel>0)
+      j->SetParallel(parallel);
    j->Quiet(quiet);
-   return j;
+   return j.borrow();
 }
 
 CMD(edit)
@@ -2474,13 +2512,59 @@ CMD(subsh)
    return e;
 }
 
+CMD(mmv)
+{
+   static const struct option mmv_opts[]=
+   {
+      {"target-directory",required_argument,0,'O'},
+      {"destination-directory",required_argument,0,'O'},
+      {"remove-target-first",no_argument,0,'e'},
+      {0}
+   };
+
+   bool remove_target=false;
+   const char *target_dir=0;
+   args->rewind();
+   int opt;
+   while((opt=args->getopt_long("eO:t:",mmv_opts,0))!=EOF)
+   {
+      switch(opt)
+      {
+      case('e'):
+	 remove_target=true;
+	 break;
+      case('t'):
+      case('O'):
+	 target_dir=optarg;
+	 break;
+      case('?'):
+      help:
+	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
+	 return 0;
+      }
+   }
+   if(!target_dir && args->count()>=3) {
+      target_dir=args->getarg(args->count()-1);
+      target_dir=alloca_strdup(target_dir);
+      args->delarg(args->count()-1);
+   }
+   if(!target_dir || args->getindex()>=args->count()) {
+      eprintf(_("Usage: %s [OPTS] <files> <target-dir>\n"),args->a0());
+      goto help;
+   }
+   mmvJob *j=new mmvJob(session->Clone(),args,target_dir,FA::RENAME);
+   if(remove_target)
+      j->RemoveTargetFirst();
+   return j;
+}
+
 CMD(mv)
 {
-   if(args->count()!=3)
+   if(args->count()!=3
+   || (args->count()==3 && last_char(args->getarg(2))=='/'))
    {
-      // xgettext:c-format
-      eprintf(_("Usage: mv <file1> <file2>\n"));
-      return 0;
+      args->setarg(0,"mmv");
+      return cmd_mmv(parent);
    }
    Job *j=new mvJob(session->Clone(),args->getarg(1),args->getarg(2));
    return j;
@@ -3057,7 +3141,7 @@ CMD(du)
    bool print_totals=false;
    bool all_files=false;
    bool file_count=false;
-   const char *exclude=0;
+   Ref<PatternSet> exclude;
    int human_opts=0;
 
    exit_code=1;
@@ -3120,7 +3204,9 @@ CMD(du)
 	 }
 	 break;
       case OPT_EXCLUDE:
-	 exclude=optarg;
+	 if(!exclude)
+	    exclude=new PatternSet();
+	 exclude->Add(PatternSet::EXCLUDE,new PatternSet::Glob(optarg));
 	 break;
       case '?':
       default:
@@ -3170,11 +3256,7 @@ CMD(du)
    if(separate_dirs && maxdepth != -1)
       j->set_maxdepth(maxdepth);
    if(exclude)
-   {
-      PatternSet *p=new PatternSet();
-      p->Add(p->EXCLUDE,new PatternSet::Glob(exclude));
-      j->SetExclude(p);
-   }
+      j->SetExclude(exclude.borrow());
    return j;
 }
 
@@ -3326,6 +3408,7 @@ CMD(get1)
       {"output",required_argument,0,'o'},
       {"remove-source-later",no_argument,0,'E'},
       {"remove-target-first",no_argument,0,'e'},
+      {"make-target-dir",no_argument,0,'d'},
       {"quiet",no_argument,0,'q'},
       {0,0,0,0}
    };
@@ -3335,11 +3418,12 @@ CMD(get1)
    bool cont=false;
    bool ascii=false;
    bool quiet=false;
+   bool do_mkdir=false;
    long long source_region_begin=0,source_region_end=FILE_END;
    long long target_region_begin=0,target_region_end=FILE_END;
    int n,p;
 
-   while((opt=args->getopt_long("arco:",get1_options))!=EOF)
+   while((opt=args->getopt_long("arco:d",get1_options))!=EOF)
    {
       switch(opt)
       {
@@ -3373,6 +3457,9 @@ CMD(get1)
       case('q'):
 	 quiet=true;
 	 break;
+      case('d'):
+	 do_mkdir=true;
+	 break;
       case '?':
       usage:
 	 eprintf(_("Usage: %s [OPTS] file\n"),args->a0());
@@ -3425,6 +3512,8 @@ CMD(get1)
       }
    }
 
+   dst=alloca_strdup(dst); // save tmp xstring
+
    FileCopyPeer *src_peer=0;
    FileCopyPeer *dst_peer=0;
 
@@ -3439,6 +3528,8 @@ CMD(get1)
    dst_peer->AutoRename(auto_rename && ResMgr::QueryBool("xfer:auto-rename",0));
    if(!cont && (target_region_begin>0 || target_region_end!=FILE_END))
       dst_peer->SetRange(target_region_begin,target_region_end);
+   if(do_mkdir)
+      dst_peer->MakeTargetDir();
 
    FileCopy *c=FileCopy::New(src_peer,dst_peer,cont);
 
